#ifndef __gui__h
#define __gui__h

#include <boost/shared_ptr.hpp>
#include <vector>

namespace gui
{

// base for GUI classes like windows, listboxes, etc.
//==================================================
template< typename Tr >
class window
{
public:
	typedef Tr                       Traits;
	typedef Traits::general::Hnd     Hnd;


public:
	inline window() { m_hnd = NULL; }
	virtual ~window() 
	{ 
		destroy(); 
	}

	inline void destroy()
	{
		if( !m_hnd ) return;
		Traits::general::destroy( m_hnd );
		m_hnd = NULL;
	}

	//general
	inline void   set_hnd    ( Hnd h ) { m_hnd = h; }
	inline Hnd    get_hnd    () const { return m_hnd; }

	inline bool   is_valid   ( ) { return Traits::general::is_valid( m_hnd ); }
	inline void   set_show   ( bool show ) { Traits::general::set_show(m_hnd, show); }

	inline void   update()  { Traits::general::update(m_hnd); }

	inline void   set_title( std::string& str ) { Traits::general::set_title( m_hnd, str ); }

	inline operator Hnd() const { return m_hnd; }

protected:
	Hnd m_hnd;
};


template <typename T> class viewdata;

//the 'view' class is the base for all GUI elements
//the element doesn't have to be a separate window.
//For example individual list control cells are views
template <typename T>
class view
{
public:
	virtual ~view()
	{
		if( m_vd )
			m_vd->remove_client( this );
	}

	//get/set view data
	inline T& get_data() { return m_vd->get_data(); }

	inline connect( view<T>* v )
	{
		init_vdata();
		m_vd->add_client(v);
		v->m_vd = m_vd;
		v->update();
	}

	view<T>& operator <<( const T& d )
	{
		init_vdata();
		*m_vd  = d;
		return *this;
	}

	virtual void update() = 0;

protected:
	boost::shared_ptr< viewdata<T> > m_vd;  //view data

	inline void init_vdata()
	{
		if( !m_vd ) 
		{
			boost::shared_ptr< viewdata<T> > s( new viewdata<T> );
			m_vd = s;
			m_vd->add_client( this );
		}
	}
};

//the 'viewdata' class provides a connection between 
//'real-world' data and views
template <typename T>
class viewdata
{
public:
	T   m_data;  //real-world data

	explicit viewdata()
	{
	}

	virtual ~viewdata()
	{
	}


	viewdata<T>& operator=( const T& d )
	{
		m_data = d;
		//the data is changing, update all clients
		std::vector< view<T>* >::iterator it;
		for( it = m_clients.begin(); it < m_clients.end(); ++it )
		{
			(*it)->update();
		}
		return *this;
	}

	inline T& get_data() { return m_data; }

	void remove_client( view<T>* v )
	{
		//the data is changing, update all clients
		std::vector< view<T>* >::iterator it;
		for( it = m_clients.begin(); it < m_clients.end(); ++it )
		{
			if( *it == v )
			{
				m_clients.erase( it );
				break;
			}
		}
	}

	inline void add_client( view<T>* v )
	{
		m_clients.push_back( v );
	}

protected:
	std::vector< view<T>* > m_clients;  //views attached to this data
};

//base for simple controls
template < typename T, typename Traits >
class vcontrol : public window<Traits>, public view<T>
{
public:

	virtual ~vcontrol() 
	{
	}
	//some of the control's GDI properties
//	boost::shared_ptr< Traits::font >   m_font;
//	boost::shared_ptr< Traits::bitmap > m_bkg;

	virtual void update()
	{
//		window<Traits>::set_font( m_font );
//		window<Traits>::set_background( m_bkg );
	}

	/*
	//set new font
	inline vcontrol<T,Traits>& operator<<( const boost::shared_ptr<Traits::font>& font )
	{
		m_font = font;
		update();
		return *this;
	}

	//set new background
	inline vcontrol<T,Traits>& operator<<( const hared_ptr<Traits::bitmap>& bmp )
	{
		m_bkg = bmp;
		update();
		return *this;
	}
	*/
};

//static control
template < typename T, typename F = std::string (*)(T&), typename Traits = win32 >
class vstatic : public vcontrol< T, Traits >
{
public:
	std::string m_text;
	F m_format; //render format (regex++, etc.)

	vstatic( F f = NULL ) { m_format = f; }

	virtual ~vstatic()
	{
	}


	virtual void update()
	{
		vcontrol<T, Traits>::update();
		m_text = m_format( get_data() );
		window<Traits>::set_title( m_text );
	}
};





//++++++++++++++++++++++++++++++++++++++++++++++++++++
//  Win32 traits
//++++++++++++++++++++++++++++++++++++++++++++++++++++
class winres
{
public:
	HINSTANCE m_hinst;
	LPTSTR    m_res;

	inline winres( HINSTANCE hinst, LPTSTR res )
	{
		m_hinst = hinst;
		m_res   = res;
	}

	inline winres( HINSTANCE hinst, int res )
	{
		m_hinst = hinst;
		m_res   = MAKEINTRESOURCE(res);
	}
};

class win32
{
public:
	//general
	typedef HWND        WinHnd;
	typedef HGDIOBJ     GdiHnd;
	typedef HBITMAP     bitmapHnd;
	typedef HIMAGELIST  imagelistHnd;
	typedef SIZE        Size;
	typedef POINT       Point;
	typedef RECT        Rect;
	typedef winres      Resource;
	typedef COLORREF    Color;


	class general
	{
	public:
		typedef WinHnd Hnd;

		static bool  is_valid ( Hnd h ) { return ::IsWindow( h )? true:false; }
		static void  set_show ( Hnd h, bool show ) { if(is_valid(h)) ::ShowWindow( h, show?SW_SHOW:SW_HIDE ); }
		static void  destroy  ( Hnd h ) { if(is_valid(h)) ::DestroyWindow( h ); }
		static void  update( Hnd h ) { if(is_valid(h)) ::InvalidateRect( h, NULL, TRUE ); }
		static void  set_title( Hnd h, std::string& txt ) { ::SetWindowText( h, txt.c_str() ); }
	};
};


};  //gui


#endif
